import math
import numpy as np

try:
    from . import ghkss_cpp
    from .ghkss_cpp import verbosity_none, verbosity_info, verbosity_debug, verbosity_trace
except ImportError:
    print("Could not import C++ extension. Please make sure that the ghkss_cpp module is compiled.")
    raise


class FilterConfig(ghkss_cpp.GhkssConfig):
    _ghkss_config_field_names = [f for f in ghkss_cpp.GhkssConfig().__dir__() if not f.startswith('_')]
    _config_fields = dict(
        batch_size=float('inf'),
        iterations=10,
    )
    __slots__ = tuple(_config_fields.keys())

    def __init__(self, **kwargs):
        super().__init__()
        for key, default in self._config_fields.items():
            setattr(self, key, default)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def _as_base(self):
        result = ghkss_cpp.GhkssConfig()
        for key in self._ghkss_config_field_names:
            setattr(result, key, getattr(self, key))
        return result

    def as_dict(self):
        return {**{k:getattr(self, k) for k in self._config_fields}, **{k:getattr(self, k) for k in self._ghkss_config_field_names}}

    def __getstate__(self):
        return self.as_dict()

    def __setstate__(self, state):
        self.__init__(**state)

    def __eq__(self, other):
        return self.as_dict() == other.as_dict()

    def replace(self, **kwargs):
        return self.__class__(**{**self.as_dict(), **kwargs})

    def set_delay_vector_pattern(self, delay_vector_timesteps=5, delay_vector_delta=1, signal_components=1):
        delay_vector_pattern = []
        for time_step in range(delay_vector_timesteps):
            for component in range(signal_components):
                delay_vector_pattern.append(time_step * signal_components * delay_vector_delta + component)
        self.delay_vector_pattern = delay_vector_pattern
        self.delay_vector_alignment = signal_components


def filter_ghkss(time_series, filter_config, return_neighbour_statistics=False):

    time_series = np.array(time_series)
    if len(time_series.shape) == 1:
        time_series = time_series.reshape(-1, 1)
    elif len(time_series.shape) != 2:
        raise ValueError("Only 1D or 2D time series are supported.")

    assert len(time_series) > 1, "There is only one time step in the time series. Maybe you need to transpose the input?"


    ghkss_config = filter_config._as_base()
    neighbour_statistics = []
    for iteration in range(filter_config.iterations):
        if filter_config.verbosity >= verbosity_info:
            print(f"Starting filter iteration {iteration+1} of {filter_config.iterations}.")
        batch_start = 0
        while batch_start < len(time_series):
            batch_end = min(batch_start + filter_config.batch_size, len(time_series))
            if filter_config.verbosity >= verbosity_info and filter_config.batch_size < float('inf'):
                print(f"Processing batch {batch_start+1} of {math.ceil(len(time_series)/filter_config.batch_size)}.")
            batch = time_series[batch_start:batch_end,:]
            filtered_batch = ghkss_cpp.filter_ghkss(batch.flatten(order='C'), ghkss_config, return_neighbour_statistics)
            if return_neighbour_statistics:
                filtered_batch, statistics = filtered_batch
                neighbour_statistics.append(statistics)
                if filter_config.verbosity >= verbosity_info:
                    print(f"Average neighbour count: {statistics.average_neighbour_count:.1f} (minimum: {statistics.minimum_neighbour_count}, maximum: {statistics.maximum_neighbour_count}).")
            filtered_batch = np.array(filtered_batch).reshape(batch.shape, order='C')
            time_series[batch_start:batch_end,:] = filtered_batch
            batch_start = batch_end

    if filter_config.verbosity >= verbosity_info:
        print(f"Filtering complete.")

    if return_neighbour_statistics:
        return time_series, neighbour_statistics
    else:
        return time_series




import numpy as np
import scipy

def simulate_lorenz(num_steps = 50000, burn_in=10000):
    def lorenz_dot(xyz, *, s=10, r=28, b=2.667):
        """
        Parameters
        ----------
        xyz : array-like, shape (3,)
           Point of interest in three-dimensional space.
        s, r, b : float
           Parameters defining the Lorenz attractor.

        Returns
        -------
        xyz_dot : array, shape (3,)
           Values of the Lorenz attractor's partial derivatives at *xyz*.
        """
        x, y, z = xyz
        x_dot = s*(y - x)
        y_dot = r*x - y - x*z
        z_dot = x*y - b*z
        return np.array([x_dot, y_dot, z_dot])

    DT = 0.01

    time_points = np.array([i*DT for i in range(burn_in,num_steps+burn_in)])

    current = np.array([0.1, 0.1, 0.1])
    integrated = scipy.integrate.solve_ivp(lambda t,y:lorenz_dot(y), y0=current, t_span=(0, time_points[-1]), t_eval=time_points, max_step=1e-2, rtol=1e-3)

    result = np.zeros((num_steps, 3))
    result[:,0] = integrated.y[0,:]
    result[:,1] = integrated.y[1,:]
    result[:,2] = integrated.y[2,:]
    #result[:,3] = np.array([i*DT for i in range(num_steps)])

    return result

def add_white_noise(data, signal_to_noise_ratio):
    noise = np.random.normal(0, 1, data.shape)
    signal_power = np.mean(np.square(data - np.mean(data, axis=0)))
    noise_power = np.mean(np.square(noise - np.mean(noise, axis=0)))
    desired_noise_power = signal_power * 10**(-signal_to_noise_ratio / 10)
    scaling_factor = np.sqrt(desired_noise_power / noise_power)
    return data + scaling_factor * noise


import os
import sys
import numpy as np

try:
    import ghkss
except ImportError:
    SCRIPT_DIR=os.path.dirname(os.path.realpath(__file__))
    sys.path.append(os.path.join(SCRIPT_DIR, '../src'))
    import ghkss



def test_knn_finder():

    np.random.seed(0)

    for components, sequence_length, delay_pattern in [
        (1, 10000, [0,1,2,3,4]),
        (1, 10000, [4,2,0,7,3]),
        (3, 10000, [1,2,3,11,12,13,21,22,23]),
    ]:
        for minimum_neighbour_count, epsilon, maximum_neighbour_count in [
            (0, -1, 10),
            (1, -1, 10),
            (0, 0.5, 50),
        ]:
            for euclidean_norm in [True, False]:

                print(f"Running test for {components=}, {sequence_length=}, {delay_pattern=}, {minimum_neighbour_count=}, {epsilon=}, {maximum_neighbour_count=}, {euclidean_norm=}.")

                time_series = np.random.normal(0, 1, (sequence_length, components)).flatten()

                knn_finder = ghkss.ghkss_cpp.KNearestNeighbourFinder(time_series, delay_vector_pattern=delay_pattern, delay_vector_alignment=components)

                for index in range(0, sequence_length-max(delay_pattern), components):
                    neighbour_indices = knn_finder.find_nearest_neighbours(index=index, minimum_neighbour_count=minimum_neighbour_count, neighbour_epsilon=epsilon, maximum_neighbour_count=maximum_neighbour_count, euclidean_norm=euclidean_norm)

                    try:
                        knn_finder.assert_knn_result(index, neighbour_indices, minimum_neighbour_count=minimum_neighbour_count, neighbour_epsilon=epsilon, maximum_neighbour_count=maximum_neighbour_count, euclidean_norm=euclidean_norm)
                    except:
                        print(f"Failed for index {index}. The neighbour set was {neighbour_indices}.")
                        raise




import numpy as np

from lorenz import simulate_lorenz, add_white_noise


print("Creating time series...")
ground_truth = simulate_lorenz()
time_series = add_white_noise(ground_truth, signal_to_noise_ratio=10)

np.savetxt("lorenz.csv", time_series, delimiter=",")



import os
import sys
import numpy as np
import math
import matplotlib.pyplot as plt

# Add the package to the path in case we're still in the source tree instead of using an installed package.
SCRIPT_DIR=os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(SCRIPT_DIR, '../src'))

import ghkss
from lorenz import simulate_lorenz, add_white_noise

def print_statistics(time_series, ground_truth):
    noise = time_series - ground_truth
    signal_power = np.mean(np.square(ground_truth - np.mean(ground_truth, axis=0)))
    noise_power = np.mean(np.square(noise - np.mean(noise, axis=0)))
    signal_to_noise_ratio = 10 * math.log10(signal_power / noise_power)
    rmse = np.sqrt(np.mean(np.square(noise), axis=0))

    print("Standard deviation of noise:", np.std(noise, axis=0))
    print("RMSE:", rmse)
    print(f"SNR: {signal_to_noise_ratio:.2f} dB")
    print()



print("Creating time series...")
ground_truth = simulate_lorenz()
time_series = add_white_noise(ground_truth, signal_to_noise_ratio=10)

# Configure the filter
filter_config = ghkss.FilterConfig()
filter_config.set_delay_vector_pattern(signal_components=3, delay_vector_timesteps=5)
filter_config.projection_dimension = 3
filter_config.iterations = 3
filter_config.neighbour_epsilon = 10
filter_config.verbosity = ghkss.verbosity_info

print("Noisy time series metrics:")
print_statistics(time_series, ground_truth)

print("Filtering time series...")
filtered_time_series, neighbour_statistics = ghkss.filter_ghkss(time_series, filter_config, return_neighbour_statistics=True)

for index, statistics in enumerate(neighbour_statistics):
    print(f"The average neighbour count in round {index+1} was {statistics.average_neighbour_count:.1f}.")

print("Filtered time series metrics:")
print_statistics(filtered_time_series, ground_truth)

# Plot the results for visual comparison
figure = plt.figure()
axis = figure.add_subplot(131, projection='3d')
axis.scatter(time_series[:,0], time_series[:,1], time_series[:,2], s=0.1)
axis.set_title("Noisy time series")
axis = figure.add_subplot(132, projection='3d')
axis.scatter(filtered_time_series[:,0], filtered_time_series[:,1], filtered_time_series[:,2], s=0.1)
axis.set_title("Filtered time series")
axis = figure.add_subplot(133, projection='3d')
axis.scatter(ground_truth[:,0], ground_truth[:,1], ground_truth[:,2], s=0.1)
axis.set_title("Ground truth")

print("Done!")
plt.show()


import math
import os
import sys
import numpy as np

try:
    import ghkss
except ImportError:
    SCRIPT_DIR=os.path.dirname(os.path.realpath(__file__))
    sys.path.append(os.path.join(SCRIPT_DIR, '../src'))
    import ghkss

try:
    import lorenz
except ImportError:
    SCRIPT_DIR=os.path.dirname(os.path.realpath(__file__))
    sys.path.append(os.path.join(SCRIPT_DIR, '../demo'))
    import lorenz


def filter_ghkss_python(time_series, config):

    filtered_time_series_cpp = ghkss.ghkss_cpp.filter_ghkss(time_series, config._as_base())

    if not hasattr(ghkss.ghkss_cpp, '_debug_info'):
        print("Warning: the C++ extension is not a debug build. Only testing the final result, not verifying intermediate steps.")

    # Check the the C++ code got the time series correctly.
    if hasattr(ghkss.ghkss_cpp, '_debug_info'):
        assert np.allclose(ghkss.ghkss_cpp._debug_info.time_series, time_series), f"original time series differ by {np.max(np.abs(filtered_time_series_cpp - time_series))}"

    # Work out the limit on the delay vector index.
    max_delay_index = len(time_series) - 1 - max(config.delay_vector_pattern)

    # Verify that the C++ code got the number of delay vectors correct.
    if hasattr(ghkss.ghkss_cpp, '_debug_info'):
        delay_vector_count = len(list(range(0,max_delay_index+1,config.delay_vector_alignment)))
        assert ghkss.ghkss_cpp._debug_info.delay_vector_count == delay_vector_count, f"The number of delay vectors doesn't match. {ghkss.ghkss_cpp._debug_info.delay_vector_count=} != {delay_vector_count=}"

    # The neighbour-finder is tested separately, so we can use it here without validating its results.
    knn_finder = ghkss.ghkss_cpp.KNearestNeighbourFinder(time_series, delay_vector_pattern=config.delay_vector_pattern, delay_vector_alignment=config.delay_vector_alignment)

    def get_neighbour_indices(index):
        if config.tisean_epsilon_widening:
            neighbours = []
            epsilon = config.neighbour_epsilon
            while len(neighbours) < config.minimum_neighbour_count:
                neighbours = knn_finder.find_nearest_neighbours(
                    index=index,
                    minimum_neighbour_count=0,
                    neighbour_epsilon=epsilon,
                    maximum_neighbour_count=config.maximum_neighbour_count,
                    euclidean_norm=config.euclidean_norm
                )
                epsilon *= math.sqrt(2)
        else:
            neighbours = knn_finder.find_nearest_neighbours(
                index=index,
                minimum_neighbour_count=config.minimum_neighbour_count,
                neighbour_epsilon=config.neighbour_epsilon,
                maximum_neighbour_count=config.maximum_neighbour_count,
                euclidean_norm=config.euclidean_norm
            )

        return neighbours


    # Build the list of all neighbour indices.
    # We get it as dict because the index is a multiple of the alignment.
    all_neighbours = {index: get_neighbour_indices(index) for index in range(0,max_delay_index+1,config.delay_vector_alignment)}

    # Check that the neighbour indices are correctly aligned
    for neighbours in all_neighbours.values():
        for neighbour in neighbours:
            assert neighbour % config.delay_vector_alignment == 0

    if hasattr(ghkss.ghkss_cpp, '_debug_info'):
        # The C++ code stores the neighbour indices in a flat array, so we need to adjust for the alignment.
        assert {i*config.delay_vector_alignment:n for i,n in enumerate(ghkss.ghkss_cpp._debug_info.all_neighbours)} == all_neighbours, "The neighbour indices don't match."



    # Create the weights
    delay_vector_weights = [1] * len(config.delay_vector_pattern)
    for offset in range(config.delay_vector_alignment):
        relevant_indices = [i for i,v in enumerate(config.delay_vector_pattern) if v % config.delay_vector_alignment == offset]
        if relevant_indices:
            delay_vector_weights[relevant_indices[0]] = 1000
            delay_vector_weights[relevant_indices[-1]] = 1000

    if hasattr(ghkss.ghkss_cpp, '_debug_info'):
        assert np.allclose(ghkss.ghkss_cpp._debug_info.delay_vector_weights, delay_vector_weights), f"The weights don't match. {delay_vector_weights=} != {ghkss.ghkss_cpp._debug_info.delay_vector_weights=}"

    # Create the traces
    traces = [0.] * config.delay_vector_alignment
    for index, weight in enumerate(delay_vector_weights):
        traces[config.delay_vector_pattern[index] % config.delay_vector_alignment] += 1. / weight

    if hasattr(ghkss.ghkss_cpp, '_debug_info'):
        assert np.allclose(ghkss.ghkss_cpp._debug_info.traces, traces), f"The traces don't match. {traces=} != {ghkss.ghkss_cpp._debug_info.traces=}"


    def get_delay_vector(index):
        assert index % config.delay_vector_alignment == 0
        return [time_series[index + i] for i in config.delay_vector_pattern]

    def calculate_initial_correction_deltas(index):
        assert index % config.delay_vector_alignment == 0

        neighbours = [get_delay_vector(i) for i in all_neighbours[index]]
        neighbours = np.array(neighbours)
        covariance_matrix = np.cov(neighbours, rowvar=False, bias=True)

        # apply the weights to the covariance matrix
        covariance_matrix *= np.array(delay_vector_weights).reshape(-1, 1) * np.array(delay_vector_weights).reshape(1, -1)

        eigenvalues, eigenvectors = np.linalg.eigh(covariance_matrix)
        # np.linalg.eigh returns eigenvectors in ascending order of eigenvalues.
        null_space_dimension = len(config.delay_vector_pattern) - config.projection_dimension
        projection_vectors = eigenvectors.T[:null_space_dimension]

        to_project = np.array(get_delay_vector(index)) - np.mean(neighbours, axis=0)
        to_project *= delay_vector_weights # appy the weights to the delay vector
        projected = np.zeros(to_project.shape)
        for vector in projection_vectors:
            projected += vector * np.dot(to_project, vector)

        return projected / delay_vector_weights


    # Calculate the initial correction deltas.
    initial_correction_deltas = []
    for index in range(0,max_delay_index+1,config.delay_vector_alignment):
        initial_correction_deltas.append(calculate_initial_correction_deltas(index))

    if hasattr(ghkss.ghkss_cpp, '_debug_info'):
        assert np.allclose(ghkss.ghkss_cpp._debug_info.initial_correction_deltas, initial_correction_deltas), f"The correction deltas don't match. {initial_correction_deltas=} != {ghkss.ghkss_cpp._debug_info.initial_correction_deltas=}"

    # We handle the trend here. (handle_trend() in the C++ code)
    final_deltas = np.zeros(len(time_series))
    for index in range(0,max_delay_index+1,config.delay_vector_alignment):
        average_correction = np.mean(np.array([initial_correction_deltas[i//config.delay_vector_alignment] for i in all_neighbours[index]]), axis=0)

        for dimension, offset_index in enumerate(config.delay_vector_pattern):
            final_deltas[index+offset_index] += ((initial_correction_deltas[index//config.delay_vector_alignment][dimension] - average_correction[dimension]) /
                                                 (traces[offset_index % config.delay_vector_alignment] * delay_vector_weights[dimension]))

    if hasattr(ghkss.ghkss_cpp, '_debug_info'):
        assert np.allclose(ghkss.ghkss_cpp._debug_info.final_deltas, final_deltas), f"The final correction deltas don't match. {final_deltas=} != {ghkss.ghkss_cpp._debug_info.final_deltas=}"

    # Now we can apply the correction deltas to the original time series.
    filtered_time_series = time_series - final_deltas
    assert np.allclose(filtered_time_series, filtered_time_series_cpp), f"The time series differ by {np.max(np.abs(filtered_time_series - filtered_time_series_cpp))}."




def test_ghkss_filter_single_component():
    # we test the C++ module for a single component.
    time_series = lorenz.simulate_lorenz(num_steps=5000)[:,0]
    time_series = lorenz.add_white_noise(time_series, signal_to_noise_ratio=10)
    time_series = time_series.flatten(order='C')

    filter_config = ghkss.FilterConfig()
    filter_config.set_delay_vector_pattern(signal_components=1, delay_vector_timesteps=5)
    filter_config.projection_dimension = 3
    filter_config.iterations = 1
    filter_config.neighbour_epsilon = 10

    filter_ghkss_python(time_series=time_series, config=filter_config)


def test_ghkss_filter_multiple_components():
    # we test the C++ module for a multiple components.
    time_series = lorenz.simulate_lorenz(num_steps=5000)
    time_series = lorenz.add_white_noise(time_series, signal_to_noise_ratio=10)
    assert time_series.shape[1] == 3
    time_series = time_series.flatten(order='C')

    filter_config = ghkss.FilterConfig()
    filter_config.set_delay_vector_pattern(signal_components=3, delay_vector_timesteps=5)
    filter_config.projection_dimension = 3
    filter_config.iterations = 1
    filter_config.neighbour_epsilon = 10

    filter_ghkss_python(time_series=time_series, config=filter_config)

def test_ghkss_filter_iterations():
    # We test whether the python implementation handles batches and iterations correctly
    time_series = lorenz.simulate_lorenz(num_steps=5000)
    time_series = lorenz.add_white_noise(time_series, signal_to_noise_ratio=10)
    assert time_series.shape[1] == 3

    filter_config = ghkss.FilterConfig()
    filter_config.set_delay_vector_pattern(signal_components=3, delay_vector_timesteps=5)
    filter_config.projection_dimension = 3
    filter_config.iterations = 3
    filter_config.neighbour_epsilon = 10

    filtered_time_series = ghkss.filter_ghkss(time_series, filter_config)

    filter_config.iterations = 1
    iteratively_filtered_time_series = time_series
    for _ in range(3):
        iteratively_filtered_time_series = ghkss.filter_ghkss(iteratively_filtered_time_series, filter_config)

    assert np.allclose(filtered_time_series, iteratively_filtered_time_series)

def test_ghkss_filter_batch():
    # We test whether the python implementation handles batches and iterations correctly
    time_series = lorenz.simulate_lorenz(num_steps=5000)
    time_series = lorenz.add_white_noise(time_series, signal_to_noise_ratio=10)
    assert time_series.shape[1] == 3

    batch_size = 2000

    filter_config = ghkss.FilterConfig()
    filter_config.set_delay_vector_pattern(signal_components=3, delay_vector_timesteps=5)
    filter_config.projection_dimension = 3
    filter_config.iterations = 3
    filter_config.neighbour_epsilon = 10
    filter_config.batch_size = batch_size

    filtered_time_series = ghkss.filter_ghkss(time_series, filter_config)

    filter_config.batch_size = float('inf')
    for batch_start in range(0, len(time_series), batch_size):
        assert np.allclose(filtered_time_series[batch_start:batch_start+batch_size,:], ghkss.filter_ghkss(time_series[batch_start:batch_start+batch_size,:], filter_config))


